; tstH64.asm
;
; A program that demonstrates the strToH64 function.

            option     casemap:none
            include    aoalib.inc   ;AoA library + constants
            
            .const

; Program title:

            align   word
ttlStr      byte    "tstH64", 0


	.data

; Data to convert:
    
in0    byte  "0", 0
in1    byte  "1", 0
in2    byte  "22", 0
in3    byte  "333", 0
in4    byte  "4444", 0
in5    byte  "55555", 0
in6    byte  "666666", 0
in7    byte  "7777777", 0
in8    byte  "88888888", 0
in9    byte  "1000_09abcdef", 0
ina    byte  "abcd_ef01_2345_6789", 0
inb    byte  " ;:,", tab, " 12345678 ", 0
inc_   byte  "aAbBcCdD;", 0
ind    byte  "eEfF_0123", 0
ine    byte  "0000_0000_0000_aBcD_eEfF_0123", 0

	            
            .code

            include getTitle.inc
            include print.inc

strToH64	proc
	push	rsi
	call	bufToH64
	pop	rsi
	ret
strToH64	endp
		
;*************************************************
;
; bufToH64
;
; Converts a string containing hexadecimal digits
; to a numeric value.
;
; Skips over any leading delimiter characters, and
; then processes all the hex digits up to the first
; delimiter character.
;
; Delimiter characters are space, comma, period
; semicolon, colon, linefeed (newline), carriage
; return, and tab.
;
; Input:
;    RSI-	Pointer to string
;
; Outputs:
;    RAX-	Converted value, or error code if
;	there was an error.
;
;    RDI-	Points at first character beyond the
;	string of hex digits.
;
;    Carry-	Clear if no error, set if error. On
;	error, RAX contains:
;	-1 if bad first character
;	-2 if overflow
;	-3 if illegal character in string.
	
delimCode  	= 	011h
underscore 	= 	010h
illChar	=	0ffh
tab	=	00ah

bufToH64	proc
    	push	rbp
    	push	rbx
    	push	rcx
    	push	rdx
    
return    	macro	
    	pop	rdx
    	pop	rcx
    	pop	rbx
    	pop	rbp
    	ret 
    	endm
        
    	lea     rbx, classTbl	; For char classification
    
    ; Skip any leading whitespace, comma, semicolon,
    ; or colon characters in the string:
    
skipDelim:
	mov     al, [rsi]
	cmp	al, 0		; Check for End of String
	je	badFirst
	add	rsi, 1
	xlat			; Get classification
	movzx   rcx, al
	cmp     al, delimCode   ; Is it a delimiter?
	je      skipDelim
	
	; First non-delimiter character must
	; be a valid hexadecimal digit:
	
	jb      nextChar        ; Hexadecimal digit
	        
	; If the first digit is bad (including underscore), 
	; return -1:
    
badFirst:
    	mov	rax, -1         ; Return -1
    	stc		        ; Sets carry flag
    	return 

    	; Process hexadecimal and decimal digits.
    	; Unroll the character conversion loop eight times 
    	; to improve performance.
    
doHex1    	macro  
    	local	nextChar
nextChar:
	movzx   rax, byte ptr [rsi]
	add 	rsi, 1
	movzx	rdx, byte ptr [rbx][rax]
	cmp     dl, underscore
	je      nextChar
	ja      delimiter
	shl	rcx, 4
	or	rcx, rdx
	endm

doHex2	macro  
    	local	nextChar
nextChar:
	movzx   rax, byte ptr [rsi]
    	add 	rsi, 1
	movzx	rdx, byte ptr [rbx][rax]
    	cmp     dl, underscore
    	je      nextChar
    	ja      delimiter
    	cmp	rcx, rbp	; Test for overflow
    	jae	overflow
    	shl	rcx, 4
    	or	rcx, rdx
    	endm

    
nextChar:
    	doHex1		; Assume no more than 16 digits.
    	doHex1          ; If 16 digits or less, overflow
    	doHex1          ; will not occur. Note that we've
    	doHex1		; already processed the first digit.
    	doHex1
    	doHex1
    	doHex1
    	doHex1
    	doHex1
    	doHex1
    	doHex1
    	doHex1
    	doHex1
    	doHex1
    	doHex1
    
    ; If there are more than 16 digits, we need to check for
    ; overflow (overflow is not guaranteed, there could be
    ; leading zeros):
    
couldOverflow:
    	mov	rbp, 1000000000000000h	;Checks for overflow
	doHex2
	doHex2
	doHex2
	doHex2
	doHex2
	doHex2
	doHex2
	doHex2
	jmp	couldOverflow	; If more than 16 digits.
            
    ; Transfer here when encountering a legitimate
    ; delimiter character (valid end of hex string).
    
delimiter:
	cmp     dl, delimCode
	jne     illegal
	mov	rax, rcx
	clc         	; Clears carry flag
	return

    ; If the first non-hex-digit character
    ; is bad, return -3:
    
illegal:
    	mov	rax, -3
    	stc		; Sets carry flag
    	return

    
overflow:
    	mov	rax, -2
    	stc		; Sets carry flag
    	return
bufToH64    endp    

; Classify each character:
;
;  0 to 0xf-   hexadecimal character
;  0x10-       underscore
;  0x11-       delimiter character
;  0xff-       illegal character
    
classTbl	equ	this byte
    byte   delimCode   ; 0x00
    byte   illChar     ; 0x01
    byte   illChar     ; 0x02
    byte   illChar     ; 0x03
    byte   illChar     ; 0x04
    byte   illChar     ; 0x05
    byte   illChar     ; 0x06
    byte   illChar     ; 0x07
    byte   illChar     ; 0x08
    byte   delimCode   ; 0x09
    byte   delimCode   ; 0x0a
    byte   illChar     ; 0x0b
    byte   illChar     ; 0x0c
    byte   delimCode   ; 0x0d
    byte   illChar     ; 0x0e
    byte   illChar     ; 0x0f
    byte   illChar     ; 0x10
    byte   illChar     ; 0x11
    byte   illChar     ; 0x12
    byte   illChar     ; 0x13
    byte   illChar     ; 0x14
    byte   illChar     ; 0x15
    byte   illChar     ; 0x16
    byte   illChar     ; 0x17
    byte   illChar     ; 0x18
    byte   illChar     ; 0x19
    byte   illChar     ; 0x1a
    byte   illChar     ; 0x1b
    byte   illChar     ; 0x1c
    byte   illChar     ; 0x1d
    byte   illChar     ; 0x1e
    byte   illChar     ; 0x1f
    byte   delimCode   ; ' '
    byte   illChar     ; '!'
    byte   illChar     ; '"'
    byte   illChar     ; '#'
    byte   illChar     ; '$'
    byte   illChar     ; '%'
    byte   illChar     ; '&'
    byte   illChar     ; '''
    byte   illChar     ; '('
    byte   illChar     ; ')'
    byte   illChar     ; '*'
    byte   illChar     ; '+'
    byte   delimCode   ; ','
    byte   illChar     ; '-'
    byte   illChar     ; '.'
    byte   illChar     ; '/'
    byte   0           ; '0'
    byte   1           ; '1'
    byte   2           ; '2'
    byte   3           ; '3'
    byte   4           ; 'illChar'
    byte   5           ; '5'
    byte   6           ; '6'
    byte   7           ; '7'
    byte   8           ; '8'
    byte   9           ; '9'
    byte   delimCode   ; ':'
    byte   delimCode   ; ';'
    byte   illChar     ; '<'
    byte   illChar     ; '='
    byte   illChar     ; '>'
    byte   illChar     ; '?'
    byte   illChar     ; '@'
    byte   10          ; 'A'
    byte   11          ; 'B'
    byte   12          ; 'C'
    byte   13          ; 'D'
    byte   14          ; 'E'
    byte   15          ; 'F'
    byte   illChar     ; 'G'
    byte   illChar     ; 'H'
    byte   illChar     ; 'I'
    byte   illChar     ; 'J'
    byte   illChar     ; 'K'
    byte   illChar     ; 'L'
    byte   illChar     ; 'M'
    byte   illChar     ; 'N'
    byte   illChar     ; 'O'
    byte   illChar     ; 'P'
    byte   illChar     ; 'Q'
    byte   illChar     ; 'R'
    byte   illChar     ; 'S'
    byte   illChar     ; 'T'
    byte   illChar     ; 'U'
    byte   illChar     ; 'V'
    byte   illChar     ; 'W'
    byte   illChar     ; 'X'
    byte   illChar     ; 'Y'
    byte   illChar     ; 'Z'
    byte   illChar     ; '['
    byte   illChar     ; '\'
    byte   illChar     ; ']'
    byte   illChar     ; '^'
    byte   underscore  ; '_'
    byte   illChar     ; '`'
    byte   10          ; 'a'
    byte   11          ; 'b'
    byte   12          ; 'c'
    byte   13          ; 'd'
    byte   14          ; 'e'
    byte   15          ; 'f'
    byte   illChar     ; 'g'
    byte   illChar     ; 'h'
    byte   illChar     ; 'i'
    byte   illChar     ; 'j'
    byte   illChar     ; 'k'
    byte   illChar     ; 'l'
    byte   illChar     ; 'm'
    byte   illChar     ; 'n'
    byte   illChar     ; 'o'
    byte   illChar     ; 'p'
    byte   illChar     ; 'q'
    byte   illChar     ; 'r'
    byte   illChar     ; 's'
    byte   illChar     ; 't'
    byte   illChar     ; 'u'
    byte   illChar     ; 'v'
    byte   illChar     ; 'w'
    byte   illChar     ; 'x'
    byte   illChar     ; 'y'
    byte   illChar     ; 'z'
    byte   illChar     ; '{'
    byte   illChar     ; '|'
    byte   illChar     ; '}'
    byte   illChar     ; '~'
    byte   illChar     ; ''
    byte   illChar     ; 0x80
    byte   illChar     ; 0x81
    byte   illChar     ; 0x82
    byte   illChar     ; 0x83
    byte   illChar     ; 0x84
    byte   illChar     ; 0x85
    byte   illChar     ; 0x86
    byte   illChar     ; 0x87
    byte   illChar     ; 0x88
    byte   illChar     ; 0x89
    byte   illChar     ; 0x8a
    byte   illChar     ; 0x8b
    byte   illChar     ; 0x8c
    byte   illChar     ; 0x8d
    byte   illChar     ; 0x8e
    byte   illChar     ; 0x8f
    byte   illChar     ; 0x90
    byte   illChar     ; 0x91
    byte   illChar     ; 0x92
    byte   illChar     ; 0x93
    byte   illChar     ; 0x94
    byte   illChar     ; 0x95
    byte   illChar     ; 0x96
    byte   illChar     ; 0x97
    byte   illChar     ; 0x98
    byte   illChar     ; 0x99
    byte   illChar     ; 0x9a
    byte   illChar     ; 0x9b
    byte   illChar     ; 0x9c
    byte   illChar     ; 0x9d
    byte   illChar     ; 0x9e
    byte   illChar     ; 0x9f
    byte   illChar     ; 0xa0
    byte   illChar     ; 0xa1
    byte   illChar     ; 0xa2
    byte   illChar     ; 0xa3
    byte   illChar     ; 0xa4
    byte   illChar     ; 0xa5
    byte   illChar     ; 0xa6
    byte   illChar     ; 0xa7
    byte   illChar     ; 0xa8
    byte   illChar     ; 0xa9
    byte   illChar     ; 0xaa
    byte   illChar     ; 0xab
    byte   illChar     ; 0xac
    byte   illChar     ; 0xad
    byte   illChar     ; 0xae
    byte   illChar     ; 0xaf
    byte   illChar     ; 0xb0
    byte   illChar     ; 0xb1
    byte   illChar     ; 0xb2
    byte   illChar     ; 0xb3
    byte   illChar     ; 0xb4
    byte   illChar     ; 0xb5
    byte   illChar     ; 0xb6
    byte   illChar     ; 0xb7
    byte   illChar     ; 0xb8
    byte   illChar     ; 0xb9
    byte   illChar     ; 0xba
    byte   illChar     ; 0xbb
    byte   illChar     ; 0xbc
    byte   illChar     ; 0xbd
    byte   illChar     ; 0xbe
    byte   illChar     ; 0xbf
    byte   illChar     ; 0xc0
    byte   illChar     ; 0xc1
    byte   illChar     ; 0xc2
    byte   illChar     ; 0xc3
    byte   illChar     ; 0xc4
    byte   illChar     ; 0xc5
    byte   illChar     ; 0xc6
    byte   illChar     ; 0xc7
    byte   illChar     ; 0xc8
    byte   illChar     ; 0xc9
    byte   illChar     ; 0xca
    byte   illChar     ; 0xcb
    byte   illChar     ; 0xcc
    byte   illChar     ; 0xcd
    byte   illChar     ; 0xce
    byte   illChar     ; 0xcf
    byte   illChar     ; 0xd0
    byte   illChar     ; 0xd1
    byte   illChar     ; 0xd2
    byte   illChar     ; 0xd3
    byte   illChar     ; 0xd4
    byte   illChar     ; 0xd5
    byte   illChar     ; 0xd6
    byte   illChar     ; 0xd7
    byte   illChar     ; 0xd8
    byte   illChar     ; 0xd9
    byte   illChar     ; 0xda
    byte   illChar     ; 0xdb
    byte   illChar     ; 0xdc
    byte   illChar     ; 0xdd
    byte   illChar     ; 0xde
    byte   illChar     ; 0xdf
    byte   illChar     ; 0xe0
    byte   illChar     ; 0xe1
    byte   illChar     ; 0xe2
    byte   illChar     ; 0xe3
    byte   illChar     ; 0xe4
    byte   illChar     ; 0xe5
    byte   illChar     ; 0xe6
    byte   illChar     ; 0xe7
    byte   illChar     ; 0xe8
    byte   illChar     ; 0xe9
    byte   illChar     ; 0xea
    byte   illChar     ; 0xeb
    byte   illChar     ; 0xec
    byte   illChar     ; 0xed
    byte   illChar     ; 0xee
    byte   illChar     ; 0xef
    byte   illChar     ; 0xf0
    byte   illChar     ; 0xf1
    byte   illChar     ; 0xf2
    byte   illChar     ; 0xf3
    byte   illChar     ; 0xf4
    byte   illChar     ; 0xf5
    byte   illChar     ; 0xf6
    byte   illChar     ; 0xf7
    byte   illChar     ; 0xf8
    byte   illChar     ; 0xf9
    byte   illChar     ; 0xfa
    byte   illChar     ; 0xfb
    byte   illChar     ; 0xfc
    byte   illChar     ; 0xfd
    byte   illChar     ; 0xfe
    byte   illChar     ; 0xff

        
; Here is the main assembly language function.

            
            public  asmMain
asmMain     proc

; Create a fake display here:
            
            push    rbp
            mov     rbp, rsp
            sub     rsp, 56
            push    rbx
	push	rsi
	push	rdi
	
timeMac	macro	parm
	lea	rsi, parm
	call	bufToH64
	endm
	
; Time execution using the following command line:
;
; powershell -Command Measure-Command { .\tstH64 }

	mov	rbx, 100000000
timeLp:	timeMac	in0
	timeMac in1
	timeMac in2
	timeMac in3
	timeMac in4
	timeMac in5
	timeMac in6
	timeMac in7
	timeMac in8
	timeMac in9
	timeMac ina
	timeMac inb
	timeMac inc_
	timeMac ind
	timeMac ine
	dec	rbx
	jne	timeLp

	
prtResults	macro	parm
	local	gotError, done
	 
	lea	rsi, parm
	call	strToH64
	jc	gotError
		
	mov	rdx, rax
	mov	r8, rsi
	call	print
	byte	"RAX=%llx, RSI='%s'", nl, 0
	jmp	done
	
gotError:	mov	rdx, rax
	call	print
	byte	"Error! Code=%ld", nl, 0
done:
	endm
	
	prtResults in0
	prtResults in1
	prtResults in2
	prtResults in3
	prtResults in4
	prtResults in5
	prtResults in6
	prtResults in7
	prtResults in8
	prtResults in9
	prtResults ina
	prtResults inb
	prtResults inc_
	prtResults ind
	prtResults ine

            
            

allDone:    pop	rdi
	pop	rsi
	pop     rbx
            leave
            ret     ;Returns to caller
        
asmMain     endp
            end